"""
This submodule contains implementations of mathematical equations. Each function is implemented as
a pure function corresponding to one or more equations. The left hand side of the equations are
returned and the parameters of the right hand side are required as function arguments. Each
function is implemented with vectorization over all numerical parameters.
"""

from . import cigre601, ieee738, joule_heating, math, radiative_cooling, solar_angles  # noqa



import numpy as np
from scipy.constants import Stefan_Boltzmann as stefan_boltzmann_constant

from ..units import Celsius, Meter, Unitless, WattPerMeter


def compute_radiative_cooling(
    surface_temperature: Celsius,
    air_temperature: Celsius,
    conductor_diameter: Meter,
    conductor_emissivity: Unitless,
) -> WattPerMeter:
    r"""Compute the radiative cooling due to black body radiation.

    Equation (27) on page 30 of :cite:p:`cigre601`.

    Parameters
    ----------
    surface_temperature:
        :math:`T_s~\left[^\circ\text{C}\right]`. The conductor surface temperature.
    air_temperature:
        :math:`T_a~\left[^\circ\text{C}\right]`. The ambient air temperature.
    conductor_diameter:
        :math:`D~\left[\text{m}\right]`. Outer diameter of the conductor.
    conductor_emissivity:
        :math:`\epsilon_s`. The emmisivity of the conductor.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`P_r~\left[\text{W}~\text{m}^{-1}\right]`. The radiative cooling of the conductor.
    """
    sigma_B = stefan_boltzmann_constant
    pi = np.pi
    D = conductor_diameter
    T_s = surface_temperature
    T_a = air_temperature
    epsilon_s = conductor_emissivity

    return pi * D * sigma_B * epsilon_s * ((T_s + 273.15) ** 4 - (T_a + 273.15) ** 4)  # type: ignore # noqa



import numpy as np
from numba import vectorize

from ..units import Date, Degrees, Radian, Unitless


def _get_day_of_year(when: Date) -> Unitless:
    YearResolutionType = np.datetime64(1, "Y")
    DayResolutionType = np.datetime64(1, "D")

    return (when.astype(DayResolutionType) - when.astype(YearResolutionType)).astype(float) + 1


def _get_hour_of_day(when: Date) -> Unitless:
    DayResolutionType = np.datetime64(1, "D")
    HourResolutionType = np.datetime64(1, "h")
    return (when.astype(HourResolutionType) - when.astype(DayResolutionType)).astype(float)


def _get_minute_of_hour(when: Date) -> Unitless:
    HourResolutionType = np.datetime64(1, "h")
    MinuteResolutionType = np.datetime64(1, "m")

    return (when.astype(MinuteResolutionType) - when.astype(HourResolutionType)).astype(float)


def compute_hour_angle_relative_to_noon(when: Date, longitude: Degrees) -> Radian:
    r"""Compute the hour angle.

    Described in the text on p. 18 of :cite:p:`ieee738`. The hour angle is the number of hours
    from noon times 15^\circ. This means that the hour angle for 11:00 is -15^\circ, and the
    hour angle for 14:00 is 30^\circ.

    This function converts the hour angle to radians by multiplying it by \frac{\pi}{12},
    which is the same as 15^\circ.

    The hour angle is used when calculating the solar altitude.
    This function does not take into account the difference between apparent/actual
    and mean solar time, which means that the result may be up to 15 minutes from the
    correct hour angle.

    Parameters
    ----------
    when:
        The time and date.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:'\omega~\left[\text{radian}\right]`. The hour angle relative to noon.
    """
    utc_hour = _get_hour_of_day(when)
    utc_minute = _get_minute_of_hour(when)
    pi = np.pi
    # We add longitude/15 since 15 degrees of longitude increases solar hour by 1
    return np.mod((-12 + utc_hour + utc_minute / 60 + longitude / 15), 24) * (
        pi / 12
    )  # pi/12 is 15 degrees


def compute_solar_declination(
    when: Date,
) -> Radian:
    r"""Compute the solar declination

    Equation (16b) on page 18 of :cite:p:`ieee738`.

    The function takes in a numpy.datetime64 object, and uses the _get_day_of_year
    function to find the day number of the year to be used to compute the solar declination.

    Parameters
    ----------
    when:
        The time and date.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\delta~\left[\text{radian}\right]`. The declination of the earth.
    """
    N = _get_day_of_year(when)
    return np.radians((23.3) * np.sin((284 + N) * 2 * np.pi / 365))


def compute_solar_azimuth_variable(
    latitude: Degrees,
    solar_declination: Radian,
    hour_angle_relative_to_noon: Radian,
) -> Radian:
    r"""Compute the solar azimuth variable.

    Equation (17b) on page 18 of :cite:p:`ieee738`.

    Parameters
    ----------
    latitude:
        :math:`Lat~\left[\text{degrees}\right]`. The latitude of the span (center).
    solar_declination:
        :math:`\delta~\left[\text{radian}\right]`. Solar declination (-23.45 to +23.45).
    hour_angle_relative_to_noon:
        :math:'\omega~\left[\text{radian}\right]`. The hour angle relative to noon.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\chi~\left[\text{radian}\right]`. The solar azimuth variable.
    """
    Lat = np.radians(latitude)
    delta = solar_declination
    omega = hour_angle_relative_to_noon
    return np.sin(omega) / ((np.sin(Lat) * np.cos(omega) - np.cos(Lat) * np.tan(delta)))


@vectorize
def _compute_solar_azimuth_constant(
    solar_azimuth_variable: Radian, hour_angle_relative_to_noon: Radian
) -> Radian:
    chi = solar_azimuth_variable
    omega = hour_angle_relative_to_noon
    pi = np.pi
    if -pi <= omega < 0:
        if chi >= 0:
            C = 0
        elif chi < 0:
            C = pi
    elif 0 <= omega < pi:
        if chi >= 0:
            C = pi
        elif chi < 0:
            C = 2 * pi
    return C


def compute_solar_azimuth_constant(
    solar_azimuth_variable: Radian, hour_angle_relative_to_noon: Radian
) -> Radian:
    r"""Compute the solar azimuth constant.

    Table 2 on page 18 of:cite:p:`ieee738`.

    Parameters
    ----------
    solar_azimuth_variable:
        :math:`\chi~\left[\text{radian}\right]`. The solar azimuth variable.
    hour_angle_relative_to_noon:
        :math:'\omega~\left[\text{radian}\right]`. The hour angle relative to noon.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`C~\left[\text{radian}\right]`. The solar azimuth constant.

    """
    return _compute_solar_azimuth_constant(solar_azimuth_variable, hour_angle_relative_to_noon)


def compute_solar_azimuth(
    solar_azimuth_constant: Radian,
    solar_azimuth_variable: Radian,
) -> Radian:
    r"""Compute the solar azimuth.

    Equation (17a) on page 18 of :cite:p:`ieee738`.

    Parameters
    ----------
    solar_azimuth_constant:
        :math:`C~\left[\text{radian}\right]`. The solar azimuth constant.
    solar_azimuth_variable:
        :math:`\chi~\left[\text{radian}\right]`. The solar azimuth variable.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`Z_c~\left[\text{radian}\right]`. The solar azimuth.
    """
    C = solar_azimuth_constant
    chi = solar_azimuth_variable
    return C + np.arctan(chi)


def compute_sin_solar_altitude(
    latitude: Degrees,
    solar_declination: Radian,
    hour_angle_relative_to_noon: Radian,
) -> Radian:
    r"""Compute the sine of the solar altitude

    This is an alteration of equation  (16a) on page 18 of :cite:p:`ieee738`.
    :math:`sin(H_c)` is calculated instead of :math:`H_c`.

    Parameters
    ----------
    latitude:
        :math:`Lat~\left[^\circ\right]`. The latitude of the span (center).
    ssolar_declination:
        :math:`\delta~\left[\text{radian}\right]`. Solar declination (-23.45 to +23.45).
    hour_angle_relative_to_noon:
        :math:'\omega~\left[\text{radian}\right]`. The hour angle relative to noon.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\sin\left(H_c\right)`. The sine of the solar altitude.
    """
    Lat = np.radians(latitude)
    delta = solar_declination
    omega = hour_angle_relative_to_noon
    return np.cos(Lat) * np.cos(delta) * np.cos(omega) + np.sin(Lat) * np.sin(delta)


def compute_cos_solar_effective_incidence_angle(
    sin_solar_altitude: Radian,
    solar_azimuth: Radian,
    conductor_azimuth: Radian,
) -> Radian:
    r"""Compute the cosine of the effective angle of incidence of the sun rays.

    This is an alteration of equation (9) on page 13 of :cite:p:`ieee738`.
    :math:`cos(\theta)` is calculated instead of :math:`\theta`.

    Parameters
    ----------
    sin_solar_altitude:
        :math:`sin(H_c)~\left[\text{radian}~\right]`. The sin of the solar altitude.
    solar_azimuth:
        :math:`Z_c~\left[\text{radian}~\right]`. The azimuth of the sun.
    conductor_azimuth:
        :math:`Z_l~\left[\text{radian}~\right]`. The azimuth of the conductor/line.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`cos(\theta)`. The cosine of the effective angle of incidence of the sun rays.
    """
    H_c = np.arcsin(sin_solar_altitude)
    Z_c = solar_azimuth
    Z_l = conductor_azimuth
    return np.cos(H_c) * np.cos(Z_c - Z_l)



import numpy as np

from linerate.units import (
    Celsius,
    KilogramPerCubeMeter,
    KilogramPerMeterPerSecond,
    Meter,
    MeterPerSecond,
    Radian,
    SquareMeterPerSecond,
    Unitless,
    WattPerMeter,
    WattPerMeterPerCelsius,
    WattPerMeterPerKelvin,
)


def compute_air_temperature_at_boundary_layer(  # T_film
    temperature_of_conductor_surface: Celsius,
    temperature_of_ambient_air: Celsius,
) -> Celsius:
    r"""Compute the temperature at the boundary layer, which is the thin air-film surrounding
    the conductor. Equal to the average of the ambient air temperature and the conductor
    surface temperature.

    Equation (6) on page 12 of :cite:p:`ieee738`.

    Parameters
    ----------
    temperature_of_conductor_surface:
        :math:'T_s ~left[\circ\text{C}\right]`. The temperature of the surface of the conductor.
    temperature_of_ambient_air:
        :math:'T_a ~left[\circ\text{C}\right]`. The temperature of the ambient air.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`(T_s + T_s)/2~\left[^\circ\text{C}\right]`. The temperature at the boundary layer,
        the thin air-film surrounding the conductor. Equal to the average of the ambient air
        temperature and the conductor surface temperature.
    """
    T_s = temperature_of_conductor_surface
    T_a = temperature_of_ambient_air
    return (T_s + T_a) / 2


def compute_dynamic_viscosity_of_air(  # mu_f
    air_temperature_at_boundary_layer: Celsius,
) -> KilogramPerMeterPerSecond:
    r"""Approximation of the dynamic viscosity of air at a given temperature.

    Equation (13a) on page 17 of :cite:p:`ieee738`.

    Parameters
    ----------
    film_temperature:
        :math:`(T_s + T_s)/2~\left[^\circ\text{C}\right]`. The temperature at the boundary layer,
        the thin air-film surrounding the conductor. Equal to the average of the ambient air
        temperature and the conductor surface temperature.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\mu_f~\left[\text{kg}~\text{m}^{-1}~\text{s}^{-1}\right]`. The dynamic viscosity
        of air.
    """
    T_film = air_temperature_at_boundary_layer
    return (1.458e-6 * (T_film + 273) ** 1.5) / (T_film + 383.4)


def compute_kinematic_viscosity_of_air(  # nu_f
    dynamic_viscosity_of_air: KilogramPerMeterPerSecond, air_density: KilogramPerCubeMeter
) -> SquareMeterPerSecond:
    r"""Compute the kinematic viscosity of air.

    Definition in text on page 25 of :cite:p:`cigre601`.

    Parameters
    ----------
    dynamic_viscosity_of_air:
        :math:`\mu_f~\left[\text{kg}~\text{m}^{-1}~\text{s}^{-1}\right]`. The dynamic viscosity of
        air.
    air_density:
        :math:`\rho_f~\left[\text{kg}~\text{m}^{-3}\right]`. The mass density of air.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\nu_f~\left[\text{m}^2~\text{s}^{-1}\right]`. The kinematic viscosity of air.
    """
    return dynamic_viscosity_of_air / air_density


def compute_reynolds_number(
    wind_speed: MeterPerSecond,
    conductor_diameter: Meter,
    kinematic_viscosity_of_air: SquareMeterPerSecond,
) -> Unitless:
    r"""Compute the Reynolds number using the conductor diameter as characteristic length scale.

    Equation (2c) on page 10 in :cite:p:`ieee738`

    Separated out kinematic viscosity of air into its own function because that is how CIGRE601
    has done it, and that way they do the calculation the same way. Kinematic viscosity is given by

    .. math:

        \nu_f = \frac{\mu_f}{\rho_f}

    where {\mu_f} is the dynamic viscosity of air and {\rho_f} is the air density.

    The Reynolds number is a dimensionless quantity that can be used to assess if a stream is
    likely to be turbulent or not. It is given by

    .. math::

        \text{N_{Re}} = \frac{V_w D_0}{\nu_f},

    where :math:`V_w` is the flow velocity, :math:`D_0` is a *characteristic length* (in our case,
    the conductor diameter) and :math:`\nu_f` is the kinematic viscosity.

    Parameters
    ----------
    wind_speed:
        :math:`V_w~\left[\text{m}~\text{s}^{-1}\right]`. The wind speed.
    conductor_diameter:
        :math:`D_0~\left[\text{m}\right]`. Outer diameter of the conductor.
    kinematic_viscosity_of_air:
        :math:`\nu_f~\left[\text{m}^2~\text{s}^{-1}\right]`. The kinematic viscosity of air.

    """
    V_w = wind_speed
    D_0 = conductor_diameter
    nu_f = kinematic_viscosity_of_air
    return V_w * D_0 / nu_f


def compute_wind_direction_factor(  # K_angle
    angle_of_attack: Radian,
) -> Unitless:
    r"""Compute the wind direction factor.

    Equation (4a) on page 11 of :cite:p:`ieee738`.

    The wind direction factor is used to calculate forced convection.

    This angle is called Phi in CIGRE601.

    Parameters
    ----------
    angle_of_attack:
        :math:`\phi~\left[\text{radian}\right]`. The wind angle-of-attack.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math: '\text{K_{angle}'. The wind direction factor.

    """
    Phi = angle_of_attack
    return 1.194 - np.cos(Phi) + 0.194 * np.cos(2 * Phi) + 0.368 * np.sin(2 * Phi)


def compute_thermal_conductivity_of_air(  # k_f
    air_temperature_at_boundary_layer: Celsius,
) -> WattPerMeterPerCelsius:
    r"""Approximation of the thermal conductivity of air.

    Equation (15a) on page 18 of :cite:p:`ieee738`.

    Parameters
    ----------
    film_temperature:
        :math:`(T_s + T_s)/2~\left[^\circ\text{C}\right]`. The temperature at the boundary layer,
        the thin air-film surrounding the conductor. Equal to the average of the ambient air
        temperature and the conductor surface temperature.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\text{k}_{f}~\left[\text{W}~\text{m}^{-1}~^\circ\text{C}^{-1}\right]`. The thermal
        conductivity of air at the boundary layer temperature.

    """
    T_film = air_temperature_at_boundary_layer
    A = 2.424e-2
    B = 7.477e-5
    C = -4.407e-9
    return np.poly1d([C, B, A])(T_film)


def compute_forced_convection(  # q_c1 or q_c2
    wind_direction_factor: Radian,
    reynolds_number: Unitless,
    thermal_conductivity_of_air: WattPerMeterPerKelvin,
    temperature_of_conductor_surface: Celsius,
    temperature_of_ambient_air: Celsius,
) -> WattPerMeter:
    r"""Compute the forced convection.

    Equation (3a) and (3b) on page 11 of :cite:p:`ieee738`.

    According to :cite:p:`ieee738`, "Equation (3a) is correct at low winds but underestimates
    forced convection at high wind speeds. Equation (3b) is correct at high wind speeds but
    underestimates forced convection at low wind speeds. At any wind speed, this standard recommends
    calculating convective heat loss with both equations, and using the larger of the two calculated
    convection heat loss rates."

    Parameters
    ----------
    wind_direction_factor:
        :math: '\text{K_{angle}'. The wind direction factor
    reynolds_number:
        :math:`\text{N_{Re}}`. The Reynolds number.
    thermal_conductivity_of_air:
        :math:`\text{k_f}~\left[\text{W}~\text{m}^{-1}~^\circ\text{C}^{-1}\right]`. The thermal
        conductivity of air at the boundary layer temperature.
    temperature_of_conductor_surface:
        :math:'T_s ~left[\circ\text{C}\right]`. The temperature of the surface of the conductor.
    temperature_of_ambient_air:
        :math:'T_a ~left[\circ\text{C}\right]`. The temperature of the ambient air.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\text{q}_{c}`. The forced convection.

    """
    K_angle = wind_direction_factor
    N_Re = reynolds_number
    k_f = thermal_conductivity_of_air
    T_s = temperature_of_conductor_surface
    T_a = temperature_of_ambient_air

    q_c1 = K_angle * (1.01 + 1.35 * N_Re**0.52) * k_f * (T_s - T_a)
    q_c2 = K_angle * 0.754 * N_Re**0.6 * k_f * (T_s - T_a)

    if hasattr(q_c1, "__len__"):
        q_cf = []
        for i in range(len(q_c1)):
            if q_c1[i] > q_c2[i]:
                q_cf.append(q_c1[i])
            else:
                q_cf.append(q_c2[i])
        return np.array(q_cf)
    else:
        if q_c1 > q_c2:
            return q_c1
        return q_c2


def compute_air_density(  # rho_f
    air_temperature_at_boundary_layer: Celsius,
    elevation: Meter,
) -> KilogramPerCubeMeter:
    r"""Compute the air density.

    Equation (14a) on page 17 of :cite:p:`ieee738`.

    The air density at the elevation of the conductor at the temperature at the boundary layer.

    Parameters
    ----------
    film_temperature:
        :math:`(T_s + T_s)/2~\left[^\circ\text{C}\right]`. The temperature at the boundary layer,
        the thin air-film surrounding the conductor. Equal to the average of the ambient air
        temperature and the conductor surface temperature.
    elevation:
        :math:`H_e~\left[\text{m}\right]`. The elevation of the conductor.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`\rho_{f}`. The air density.

    """
    T_film = air_temperature_at_boundary_layer
    H_e = elevation
    return (np.poly1d([6.379e-9, -1.525e-4, 1.293])(H_e)) / (1 + 0.00367 * T_film)


def compute_natural_convection(  # q_cn
    air_density: KilogramPerCubeMeter,
    conductor_diameter: Meter,
    temperature_of_conductor_surface: Celsius,
    temperature_of_ambient_air: Celsius,
) -> WattPerMeter:
    r"""Compute the natural convection.

    Equation (5a) on page 12 of :cite:p:`ieee738`.

    Temperature of conductor surface must be larger than or equal to temperature of ambient air

    Parameters
    ----------
    air_density:
        :math:`\rho_{f}`. The air density.
    conductor_diameter:
        :math:`D_0~\left[\text{m}\right]`. Outer diameter of the conductor.
    temperature_of_conductor_surface:
        :math:'T_s ~left[\circ\text{C}\right]`. The temperature of the surface of the conductor.
    temperature_of_ambient_air:
        :math:'T_a ~left[\circ\text{C}\right]`. The temperature of the ambient air.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`q_{cn}~\left[\text{W}~\text{m}^{-1}\right]`. The natural convection.
    """
    rho_f = air_density
    D_0 = conductor_diameter
    T_s = temperature_of_conductor_surface
    T_a = temperature_of_ambient_air
    return 3.645 * rho_f**0.5 * D_0**0.75 * (T_s - T_a) ** 1.25


def compute_convective_cooling(
    forced_convection: WattPerMeter,
    natural_convection: WattPerMeter,
) -> WattPerMeter:
    r"""Compute the convective cooling of the conductor.

    On page 11 in :cite:p:`ieee738, it says that one should calculate both forced and natural
    convection, and choose the larger of the two as the convective cooling.

    Parameters
    ----------
    forced_convection:
        :math:`q_c. The forced convection.
    natural_convection:
        :math:`q_{cn}~\left[\text{W}~\text{m}^{-1}\right]`. The natural convection.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`q_c~\left[\text{W}~\text{m}^{-1}\right]`. The convective cooling of the conductor.
        Either equal to the forced or the natural convection, whichever is the largest.
    """
    q_cf = forced_convection
    q_cn = natural_convection

    return np.maximum(q_cf, q_cn)



"""
These thermal model classes provide an easy-to-use interface to compute the thermal rating.
They store conductor and span metadata as well as the weather parameters, compute all the
heating and cooling effects and use those to estimate the thermal rating and conductor
temperature. All numerical heavy-lifting is handled by the ``linerate.equations`` and the
``linerate.solver`` modules.
"""

from abc import ABC, abstractmethod
from numbers import Real
from typing import Dict

import numpy as np

from linerate import solver
from linerate.equations import (
    cigre601,
    ieee738,
    joule_heating,
    math,
    radiative_cooling,
    solar_angles,
)
from linerate.equations.math import switch_cos_sin
from linerate.types import Span, Weather
from linerate.units import (
    Ampere,
    Celsius,
    Date,
    JoulePerKilogramPerKelvin,
    OhmPerMeter,
    WattPerMeter,
)

__all__ = ["ThermalModel", "Cigre601", "IEEE738"]


def _copy_method_docstring(parent_class):
    def inner(func):
        func.__doc__ = getattr(parent_class, func.__name__).__doc__
        return func

    return inner


class ThermalModel(ABC):
    """Abstract class for a minimal conductor thermal model."""

    @abstractmethod
    def __init__(self, span: Span, weather: Weather):
        self.span = span
        self.weather = weather

    @abstractmethod
    def compute_resistance(self, conductor_temperature: Celsius, current: Ampere) -> OhmPerMeter:
        r"""Compute the conductor resistance, :math:`R~\left[\Omega~\text{m}^{-1}\right]`.

        Parameters
        ----------
        conductor_temperature:
            :math:`T_\text{av}~\left[^\circ\text{C}\right]`. The average conductor temperature.
        current:
            :math:`I~\left[\text{A}\right]`. The current.

        Returns
        -------
        Union[float, float64, ndarray[Any, dtype[float64]]]
            :math:`R~\left[\Omega\right]`. The resistance at the given temperature and current.
        """
        resistance = joule_heating.compute_resistance(
            conductor_temperature,
            temperature1=self.span.conductor.temperature1,
            temperature2=self.span.conductor.temperature2,
            resistance_at_temperature1=self.span.conductor.resistance_at_temperature1,
            resistance_at_temperature2=self.span.conductor.resistance_at_temperature2,
        )

        A = self.span.conductor.aluminium_cross_section_area
        b = self.span.conductor.constant_magnetic_effect
        m = self.span.conductor.current_density_proportional_magnetic_effect
        max_increase = self.span.conductor.max_magnetic_core_relative_resistance_increase

        return joule_heating.correct_resistance_acsr_magnetic_core_loss(
            ac_resistance=resistance,
            current=current,
            aluminium_cross_section_area=A,
            constant_magnetic_effect=b,
            current_density_proportional_magnetic_effect=m,
            max_relative_increase=max_increase,
        )

    @abstractmethod
    def compute_joule_heating(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        r"""Compute the Joule heating, :math:`P_J~\left[\text{W}~\text{m}^{-1}\right]`.

        Parameters
        ----------
        conductor_temperature:
            :math:`T_\text{av}~\left[^\circ\text{C}\right]`. The average conductor temperature.
        current:
            :math:`I~\left[\text{A}\right]`. The current.

        Returns
        -------
        Union[float, float64, ndarray[Any, dtype[float64]]]
            :math:`P_J~\left[\text{W}~\text{m}^{-1}\right]`. The Joule heating.
        """
        resistance = self.compute_resistance(conductor_temperature, current)
        return joule_heating.compute_joule_heating(current, resistance)

    @abstractmethod
    def compute_solar_heating(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        r"""Compute the solar heating, :math:`P_S~\left[\text{W}~\text{m}^{-1}\right]`.

        Parameters
        ----------
        conductor_temperature:
            :math:`T_\text{av}~\left[^\circ\text{C}\right]`. The average conductor temperature.
        current:
            :math:`I~\left[\text{A}\right]`. The current.

        Returns
        -------
        Union[float, float64, ndarray[Any, dtype[float64]]]
            :math:`P_s~\left[\text{W}~\text{m}^{-1}\right]`. The solar heating.
        """
        raise NotImplementedError

    @abstractmethod
    def compute_convective_cooling(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        r"""Compute the convective cooling, :math:`P_c~\left[\text{W}~\text{m}^{-1}\right]`.

        Parameters
        ----------
        conductor_temperature:
            :math:`T_\text{av}~\left[^\circ\text{C}\right]`. The average conductor temperature.
        current:
            :math:`I~\left[\text{A}\right]`. The current.

        Returns
        -------
        Union[float, float64, ndarray[Any, dtype[float64]]]
            :math:`P_c~\left[\text{W}~\text{m}^{-1}\right]`. The convective cooling.
        """
        raise NotImplementedError

    @abstractmethod
    def compute_radiative_cooling(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        r"""Compute the radiative cooling, :math:`P_r~\left[\text{W}~\text{m}^{-1}\right]`.

        Parameters
        ----------
        conductor_temperature:
            :math:`T_\text{av}~\left[^\circ\text{C}\right]`. The average conductor temperature.
        current:
            :math:`I~\left[\text{A}\right]`. The current.

        Returns
        -------
        Union[float, float64, ndarray[Any, dtype[float64]]]
            :math:`P_r~\left[\text{W}~\text{m}^{-1}\right]`. The radiative cooling heating.
        """
        return radiative_cooling.compute_radiative_cooling(
            surface_temperature=conductor_temperature,
            air_temperature=self.weather.air_temperature,
            conductor_diameter=self.span.conductor.conductor_diameter,
            conductor_emissivity=self.span.conductor.emissivity,
        )

    def compute_heat_balance(self, conductor_temperature: Celsius, current: Ampere) -> WattPerMeter:
        r"""Compute the conductor's heat balance. Positive means heating and negative means cooling.

        Parameters
        ----------
        conductor_temperature:
            :math:`T_\text{av}~\left[^\circ\text{C}\right]`. The average conductor temperature.
        current:
            :math:`I~\left[\text{A}\right]`. The current.

        Returns
        -------
        Union[float, float64, ndarray[Any, dtype[float64]]]
            :math:`P_J + P_s - P_c - P_r~\left[\text{W}~\text{m}^{-1}\right]`. The heat balance.
        """
        P_j = self.compute_joule_heating(conductor_temperature, current)
        P_s = self.compute_solar_heating(conductor_temperature, current)
        P_c = self.compute_convective_cooling(conductor_temperature, current)
        P_r = self.compute_radiative_cooling(conductor_temperature, current)
        return P_j + P_s - P_c - P_r

    def compute_info(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> Dict[str, WattPerMeter]:
        r"""Create a dictionary with the different heating and cooling effects.

        Parameters
        ----------
        conductor_temperature:
            :math:`T_\text{av}~\left[^\circ\text{C}\right]`. The average conductor temperature.
        current:
            :math:`I~\left[\text{A}\right]`. The current.

        Returns
        -------
        Dict[str, WattPerMeter]
            A dictionary with the magnitude of the different heating and cooling effects.
        """
        return {
            "convective_cooling": self.compute_convective_cooling(conductor_temperature, current),
            "radiative_cooling": self.compute_radiative_cooling(conductor_temperature, current),
            "joule_heating": self.compute_joule_heating(conductor_temperature, current),
            "solar_heating": self.compute_solar_heating(conductor_temperature, current),
        }

    def compute_steady_state_ampacity(
        self,
        max_conductor_temperature: Celsius,
        min_ampacity: Ampere = 0,
        max_ampacity: Ampere = 5000,
        tolerance: float = 1.0,
    ) -> Ampere:
        r"""Use the bisection method to compute the steady-state thermal rating (ampacity).

        Parameters
        ----------
        max_conductor_temperature:
            :math:`T_\text{max}~\left[^\circ\text{C}\right]`. Maximum allowed conductor temperature
        min_ampacity:
            :math:`I_\text{min}~\left[\text{A}\right]`. Lower bound for the numerical scheme for
            computing the ampacity
        max_ampacity:
            :math:`I_\text{min}~\left[\text{A}\right]`. Upper bound for the numerical scheme for
            computing the ampacity
        tolerance:
            :math:`\Delta I~\left[\text{A}\right]`. The numerical accuracy of the ampacity. The
            bisection iterations will stop once the numerical ampacity uncertainty is below
            :math:`\Delta I`. The bisection method will run for
            :math:`\left\lceil\frac{I_\text{min} - I_\text{min}}{\Delta I}\right\rceil` iterations.

        Returns
        -------
        Union[float, float64, ndarray[Any, dtype[float64]]]
            :math:`I~\left[\text{A}\right]`. The thermal rating.
        """
        I = solver.compute_conductor_ampacity(  # noqa
            self.compute_heat_balance,
            max_conductor_temperature=max_conductor_temperature,
            min_ampacity=min_ampacity,
            max_ampacity=max_ampacity,
            tolerance=tolerance,
        )
        n = self.span.num_conductors
        return I * n

    def compute_conductor_temperature(
        self,
        current: Ampere,
        min_temperature: Celsius = -30,
        max_temperature: Celsius = 150,
        tolerance: float = 0.5,
    ) -> Celsius:
        r"""Use the bisection method to compute the steady state conductor temperature.

        Parameters
        ----------
        current:
            :math:`I_\text{max}~\left[\text{A}\right]`. The current flowing through the conductor.
        min_temperature:
            :math:`T_\text{min}~\left[^\circ\text{C}\right]`. Lower bound for the numerical scheme
            for computing the temperature
        max_temperature:
            :math:`T_\text{max}~\left[^\circ\text{C}\right]`. Upper bound for the numerical scheme
            for computing the temperature
        tolerance:
            :math:`\Delta T~\left[^\circ\text{C}\right]`. The numerical accuracy of the
            temperature. The bisection iterations will stop once the numerical temperature
            uncertainty is below :math:`\Delta T`. The bisection method will run for
            :math:`\left\lceil\frac{T_\text{min} - T_\text{min}}{\Delta T}\right\rceil` iterations.

        Returns
        -------
        Union[float, float64, ndarray[Any, dtype[float64]]]
            :math:`I~\left[\text{A}\right]`. The thermal rating.
        """
        T = solver.compute_conductor_temperature(
            self.compute_heat_balance,
            current=current,
            min_temperature=min_temperature,
            max_temperature=max_temperature,
            tolerance=tolerance,
        )
        n = self.span.num_conductors
        return T / n


class Cigre601(ThermalModel):
    def __init__(
        self,
        span: Span,
        weather: Weather,
        time: Date,
        max_reynolds_number: Real = 4000,  # Max value of the angle correction in CIGRE601
    ):
        super().__init__(span, weather)
        self.time = time
        self.max_reynolds_number = max_reynolds_number

    @_copy_method_docstring(ThermalModel)
    def compute_resistance(self, conductor_temperature: Celsius, current: Ampere) -> OhmPerMeter:
        return super().compute_resistance(
            conductor_temperature=conductor_temperature, current=current
        )

    @_copy_method_docstring(ThermalModel)
    def compute_joule_heating(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        return super().compute_joule_heating(
            conductor_temperature=conductor_temperature, current=current
        )

    @_copy_method_docstring(ThermalModel)
    def compute_solar_heating(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        alpha_s = self.span.conductor.solar_absorptivity
        F = self.span.ground_albedo
        phi = self.span.latitude
        gamma_c = self.span.conductor_azimuth
        y = self.span.conductor_altitude
        N_s = self.weather.clearness_ratio
        D = self.span.conductor.conductor_diameter

        omega = solar_angles.compute_hour_angle_relative_to_noon(self.time, self.span.longitude)
        delta = solar_angles.compute_solar_declination(self.time)
        sin_H_s = solar_angles.compute_sin_solar_altitude(phi, delta, omega)
        chi = solar_angles.compute_solar_azimuth_variable(phi, delta, omega)
        C = solar_angles.compute_solar_azimuth_constant(chi, omega)
        gamma_s = solar_angles.compute_solar_azimuth(C, chi)  # Z_c in IEEE
        cos_eta = solar_angles.compute_cos_solar_effective_incidence_angle(
            sin_H_s, gamma_s, gamma_c
        )
        sin_eta = switch_cos_sin(cos_eta)

        I_B = cigre601.solar_heating.compute_direct_solar_radiation(sin_H_s, N_s, y)
        I_d = cigre601.solar_heating.compute_diffuse_sky_radiation(I_B, sin_H_s)
        I_T = cigre601.solar_heating.compute_global_radiation_intensity(
            I_B, I_d, F, sin_eta, sin_H_s
        )
        return cigre601.solar_heating.compute_solar_heating(
            alpha_s,
            I_T,
            D,
        )

    @_copy_method_docstring(ThermalModel)
    def compute_convective_cooling(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        D = self.span.conductor.conductor_diameter
        d = self.span.conductor.outer_layer_strand_diameter
        y = self.span.conductor_altitude
        beta = self.span.inclination
        V = self.weather.wind_speed
        T_a = self.weather.air_temperature
        T_c = conductor_temperature
        T_f = 0.5 * (T_c + T_a)

        # Compute physical quantities
        lambda_f = cigre601.convective_cooling.compute_thermal_conductivity_of_air(T_f)
        mu_f = cigre601.convective_cooling.compute_dynamic_viscosity_of_air(T_f)
        gamma_f = cigre601.convective_cooling.compute_air_density(T_f, y)
        nu_f = cigre601.convective_cooling.compute_kinematic_viscosity_of_air(mu_f, gamma_f)
        c_f: JoulePerKilogramPerKelvin = 1005
        delta = math.compute_angle_of_attack(
            self.weather.wind_direction, self.span.conductor_azimuth
        )

        # Compute unitless quantities
        Re = np.minimum(
            cigre601.convective_cooling.compute_reynolds_number(V, D, nu_f),
            self.max_reynolds_number,
        )
        Gr = cigre601.convective_cooling.compute_grashof_number(D, T_c, T_a, nu_f)
        Pr = cigre601.convective_cooling.compute_prandtl_number(lambda_f, mu_f, c_f)
        Rs = cigre601.convective_cooling.compute_conductor_roughness(D, d)

        # Compute nusselt numbers
        Nu_90 = cigre601.convective_cooling.compute_perpendicular_flow_nusseltnumber(
            reynolds_number=Re, conductor_roughness=Rs
        )
        Nu_delta = cigre601.convective_cooling.correct_wind_direction_effect_on_nusselt_number(
            Nu_90, delta, Rs
        )

        Nu_0 = cigre601.convective_cooling.compute_horizontal_natural_nusselt_number(Gr, Pr)
        Nu_beta = cigre601.convective_cooling.correct_natural_nusselt_number_inclination(
            Nu_0, beta, Rs
        )

        Nu = cigre601.convective_cooling.compute_nusselt_number(
            forced_convection_nusselt_number=Nu_delta, natural_nusselt_number=Nu_beta
        )

        return cigre601.convective_cooling.compute_convective_cooling(
            surface_temperature=conductor_temperature,
            air_temperature=self.weather.air_temperature,
            nusselt_number=Nu,
            thermal_conductivity_of_air=lambda_f,
        )

    @_copy_method_docstring(ThermalModel)
    def compute_radiative_cooling(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        return super().compute_radiative_cooling(
            conductor_temperature=conductor_temperature, current=current
        )

    def compute_temperature_gradient(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> Celsius:
        r"""Estimate the difference between the core temperature and the surface temperature.

        Parameters
        ----------
        conductor_temperature:
            :math:`T_\text{av}~\left[^\circ\text{C}\right]`. The average conductor temperature.
        current:
            :math:`I~\left[\text{A}\right]`. The current.

        Returns
        -------
        Union[float, float64, ndarray[Any, dtype[float64]]]
            :math:`T_c - T_s~\left[^\circ \text{C}\right]`. The difference between the core and the
            surface temperature of the conductor.
        """
        n = self.span.num_conductors
        T_c = conductor_temperature
        I = current / n  # noqa
        R = self.compute_resistance(conductor_temperature=T_c, current=I)
        return cigre601.convective_cooling.compute_temperature_gradient(
            total_heat_gain=I * R,
            conductor_thermal_conductivity=self.span.conductor.thermal_conductivity,  # type: ignore  # noqa
            core_diameter=self.span.conductor.core_diameter,
            conductor_diameter=self.span.conductor.conductor_diameter,
        )


class IEEE738(ThermalModel):
    def __init__(
        self,
        span: Span,
        weather: Weather,
        time: Date,
        max_reynolds_number: Real = 50_000,  # Max Reynolds number for forced convection
    ):
        super().__init__(span, weather)
        self.time = time
        self.max_reynolds_number = max_reynolds_number

    @_copy_method_docstring(ThermalModel)
    def compute_resistance(self, conductor_temperature: Celsius, current: Ampere) -> OhmPerMeter:
        return super().compute_resistance(
            conductor_temperature=conductor_temperature, current=current
        )

    @_copy_method_docstring(ThermalModel)
    def compute_joule_heating(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        return super().compute_joule_heating(
            conductor_temperature=conductor_temperature, current=current
        )

    @_copy_method_docstring(ThermalModel)
    def compute_solar_heating(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        alpha_s = self.span.conductor.solar_absorptivity  # alpha in IEEE
        phi = self.span.latitude  # Lat in IEEE
        gamma_c = self.span.conductor_azimuth  # Z_l i IEEE
        y = self.span.conductor_altitude  # H_e in IEEE
        D = self.span.conductor.conductor_diameter  # D_0 in IEEE

        omega = solar_angles.compute_hour_angle_relative_to_noon(self.time, self.span.longitude)
        delta = solar_angles.compute_solar_declination(self.time)
        sin_H_c = solar_angles.compute_sin_solar_altitude(phi, delta, omega)
        Q_s = ieee738.solar_heating.compute_total_heat_flux_density(sin_H_c, True)
        K_solar = ieee738.solar_heating.compute_solar_altitude_correction_factor(y)
        Q_se = ieee738.solar_heating.compute_elevation_correction_factor(K_solar, Q_s)
        chi = solar_angles.compute_solar_azimuth_variable(phi, delta, omega)
        C = solar_angles.compute_solar_azimuth_constant(chi, omega)
        Z_c = solar_angles.compute_solar_azimuth(C, chi)
        cos_theta = solar_angles.compute_cos_solar_effective_incidence_angle(sin_H_c, Z_c, gamma_c)

        return ieee738.solar_heating.compute_solar_heating(alpha_s, Q_se, cos_theta, D)

    @_copy_method_docstring(ThermalModel)
    def compute_convective_cooling(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        D = self.span.conductor.conductor_diameter  # D_0 in IEEE
        y = self.span.conductor_altitude  # H_e in IEEE
        V = self.weather.wind_speed  # V_w in IEEE
        T_a = self.weather.air_temperature
        T_c = conductor_temperature
        T_f = 0.5 * (T_c + T_a)  # T_film in IEEE

        mu_f = ieee738.convective_cooling.compute_dynamic_viscosity_of_air(T_f)
        rho_f = ieee738.convective_cooling.compute_air_density(T_f, y)
        nu_f = ieee738.convective_cooling.compute_kinematic_viscosity_of_air(mu_f, rho_f)
        Re = np.minimum(
            ieee738.convective_cooling.compute_reynolds_number(V, D, nu_f),  # N_Re in IEEE
            self.max_reynolds_number,
        )
        delta = math.compute_angle_of_attack(
            self.weather.wind_direction, self.span.conductor_azimuth
        )  # Phi in IEEE
        K_angle = ieee738.convective_cooling.compute_wind_direction_factor(delta)
        k_f = ieee738.convective_cooling.compute_thermal_conductivity_of_air(T_f)
        q_cf = ieee738.convective_cooling.compute_forced_convection(K_angle, Re, k_f, T_c, T_a)
        q_cn = ieee738.convective_cooling.compute_natural_convection(rho_f, D, T_c, T_a)
        return ieee738.convective_cooling.compute_convective_cooling(q_cf, q_cn)

    @_copy_method_docstring(ThermalModel)
    def compute_radiative_cooling(
        self, conductor_temperature: Celsius, current: Ampere
    ) -> WattPerMeter:
        return super().compute_radiative_cooling(
            conductor_temperature=conductor_temperature, current=current
        )



from typing import Union

try:
    from typing import Annotated
except ImportError:  # Python version <3.9
    from typing_extensions import Annotated

import numpy as np
import numpy.typing as npt

FloatOrFloatArray = Union[float, np.float64, npt.NDArray[np.float64]]
BoolOrBoolArray = Union[bool, np.bool_, npt.NDArray[np.bool_]]

OhmPerMeter = Annotated[FloatOrFloatArray, "Ω/m"]
Ampere = Annotated[FloatOrFloatArray, "A"]
Radian = Annotated[FloatOrFloatArray, "rad"]
Degrees = Annotated[FloatOrFloatArray, "°"]
Kelvin = Annotated[FloatOrFloatArray, "K"]
Celsius = Annotated[FloatOrFloatArray, "°C"]
Meter = Annotated[FloatOrFloatArray, "m"]
MeterPerSecond = Annotated[FloatOrFloatArray, "m/s"]
MeterPerSquareSecond = Annotated[FloatOrFloatArray, "m/s²"]
SquareMeterPerAmpere = Annotated[FloatOrFloatArray, "m²/A"]
SquareMeter = Annotated[FloatOrFloatArray, "m²"]
Unitless = Annotated[FloatOrFloatArray, ""]
JoulePerKilogramPerKelvin = Annotated[FloatOrFloatArray, "J/(kg K)"]
WattPerSquareMeter = Annotated[FloatOrFloatArray, "W/m²"]
WattPerMeter = Annotated[FloatOrFloatArray, "W/m"]
WattPerMeterPerKelvin = Annotated[FloatOrFloatArray, "W/(m K)"]
WattPerMeterPerCelsius = Annotated[FloatOrFloatArray, "W/(m °C)"]
SquareMeterPerSecond = Annotated[FloatOrFloatArray, "m²/s"]
KilogramPerMeterPerSecond = Annotated[FloatOrFloatArray, "kg/(m s)"]
KilogramPerCubeMeter = Annotated[FloatOrFloatArray, "kg/m³"]

Date = Union[np.datetime64, npt.NDArray[np.datetime64]]



import numpy as np

from linerate.equations.math import switch_cos_sin
from linerate.units import (
    BoolOrBoolArray,
    Meter,
    Radian,
    Unitless,
    WattPerMeter,
    WattPerSquareMeter,
)

_clear_atmosphere_polynomial = np.poly1d(
    [-4.07608e-9, 1.94318e-6, -3.61118e-4, 3.46921e-2, -1.9220, 63.8044, -42.2391]
)


_industrial_atmosphere_polynomial = np.poly1d(
    [1.3236e-8, -4.3446e-6, 5.4654e-4, -3.1658e-2, 6.6138e-1, 14.2110, 53.1821]
)


def compute_total_heat_flux_density(
    sin_solar_altitude: Radian,
    clear_atmosphere: BoolOrBoolArray,
) -> WattPerSquareMeter:
    r"""Compute the heat flux density received by a surface at sea level.

    Equation (18) on page 19 of :cite:p:`ieee738`.

    This function takes in the sin of the solar altitude, :math`H_c`, in radians.
    This is because this is what is calculated in compute_sin_solar_altitude.
    This function therefore takes the arcsin of sin_solar_altitude, and then
    converts it do degrees.

    Parameters
    ----------
    sin_solar_altitude:
        :math:`sin(H_c)~\left[\text{radian}~\right]`. The sin of the solar altitude.
    clear_atmosphere:
        True or False. True: clear atmosphere. False: industrial atmosphere.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`Q_s~\left[\text{W}~\text{m}^{-2}\right]`
    """
    sin_H_c = sin_solar_altitude
    H_c = np.degrees(np.arcsin(sin_H_c))
    Q_S = np.where(
        clear_atmosphere,
        _clear_atmosphere_polynomial(H_c),
        _industrial_atmosphere_polynomial(H_c),
    )
    return np.where(Q_S >= 0, Q_S, 0 * Q_S)


def compute_solar_altitude_correction_factor(
    height_above_sea_level_of_conductor: Meter,
) -> Unitless:
    r"""Compute the solar altitude correction factor.

    Equation (20) on page 20 of :cite:p:`ieee738`.

    Parameters
    ----------
    height_above_sea_level_of_conductor:
        :math:`H_e~\left[\text{m}~\right]`. The elevation of the conductor.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`K_{solar}`
    """
    H_e = height_above_sea_level_of_conductor
    A = 1
    B = 1.148e-4
    C = -1.108e-8
    return np.poly1d([C, B, A])(H_e)


def compute_elevation_correction_factor(
    solar_altitude_correction_factor: Unitless,
    total_heat_flux_density: WattPerSquareMeter,
) -> WattPerSquareMeter:
    r"""Compute the elevation correction factor.

    Equation (19) on page 19 of :cite:p:`ieee738`.

    The equation is used to correct the solar heat intensity for altitude.

    Parameters
    ----------
    solar_altitude_correction_factor:
        :math:`K_{solar}\left[ \right]`
    total_heat_flux_density:
        :math:`Q_s~\left[\text{W}~\text{m}^{-2}\right]`

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`Q_{se}~\left[\text{W}~\text{m}^{-2}\right]`. The elevation correction factor.
    """
    K_solar = solar_altitude_correction_factor
    Q_s = total_heat_flux_density
    return K_solar * Q_s


def compute_solar_heating(
    absorptivity: Unitless,
    elevation_correction_factor: WattPerSquareMeter,
    cos_solar_effective_incidence_angle: Radian,
    projected_area_of_conductor: Meter,  # Meter squared per linear meter
) -> WattPerMeter:
    r"""Compute the solar heating experienced by the ocnductor.

    Equation (8) on page 9 of :cite:p:`ieee738`.

    Parameters
    ----------
    absorptivity:
        :math:`\alpha`. Material constant. According to :cite:p:`ieee738`, it has a range from
        0.23 to 0.91, with new conductors having a value between 0.2 and 0.4, and over time
        increasing to between 0.5 and 0.9.
    elevation_correction_factor:
        :math:`Q_{se}~\left[\text{W}~\text{m}^{-2}\right]`.The elevation correction factor.
    cos_solar_effective_incidence_angle:
        :math:`cos(\theta)~\left[\text{radian}\right]`. The cosine of the effective angle of
        incidence of the sun’s rays.
    projected_area_of_conductor:
        :math:`A~\left[\text{m}\right]`. :math:`\text{m}^2` per linear m. Equal to the outer
        diameter of the conductor.

    Returns
    -------
    Union[float, float64, ndarray[Any, dtype[float64]]]
        :math:`P_S~\left[\text{W}~\text{m}^{-1}\right]`. The solar heating of the conductor.
    """
    alpha = absorptivity
    Q_se = elevation_correction_factor
    sin_theta = switch_cos_sin(cos_solar_effective_incidence_angle)
    A = projected_area_of_conductor
    return alpha * Q_se * sin_theta * A
